#!/usr/bin/perl

######################################################################
package Cognition::uF::XFN;
######################################################################

use XML::LibXML qw(:all);
use strict;
use CGI::Util;
use Data::Dumper;
use Cognition::MagicString;
use Cognition::Export;

BEGIN
{
	our $VERSION = 0.1;
	our @ISA = qw(Cognition::uF);
}

sub consume
{
	my $page        = shift;
	my $pageURL     = $page->abs_url;
	my $pageSubject = $page->{RDF}->{$pageURL};
	my $version;
	
	if ($page->{Profiles}->{'http://gmpg.org/xfn/11'})
	{
		$version = 11;
	}
	elsif ($page->{Profiles}->{'http://gmpg.org/xfn/1'})
	{
		$version = 10;
	}
	elsif (!$page->{options}->{uf_strict_profiles})
	{
		$version = 11;
	}
	else
	{
		return;
	}
		
	my %R = (
		'contact'       => ':',
		'acquaintance'  => 'K:',
		'friend'        => 'K:',
		'met'           => 'SK:',
		'co-worker'     => 'S:',
		'colleague'     => 'S:',
		'co-resident'   => 'SKT:',
		'neighbor'      => 'S:',
		'child'         => 'I:parent',
		'parent'        => 'I:child',
		'sibling'       => 'S:',
		'spouse'        => 'SK:',
		'kin'           => 'S:',
		'muse'          => ':',
		'crush'         => 'K:',
		'date'          => 'SK:',
		'sweetheart'    => 'SK:',
		'me'            => 'S:'
	);
	
	if ($version < 11)
	{
		delete $R{'contact'};
		delete $R{'kin'};
		delete $R{'me'};
	}
	
	if ($page->{Profiles}->{'http://xen.adactio.com/'})
	{
		$R{'nemesis'}    = 'SK:';
		$R{'enemy'}      = 'K:';
		$R{'nuisance'}   = 'K:';
		$R{'evil-twin'}  = 'S:';
		$R{'rival'}      = 'K:';
		$R{'fury'}       = ':';
		$R{'creep'}      = ':';
	}

	my $regexp = join '|', (keys %R);
	$regexp = "(^|\\s)($regexp)(\\s|\$)";
	
	my $needToMergeSubject;
	my $representative = findSubject($page->{RDF},
		firstOf($pageSubject,
			'http://purl.org/uF/hCard/terms/representative',
			'http://xmlns.com/foaf/0.1/primaryTopic',
			'http://xmlns.com/foaf/0.1/topic'));
	
	my @links = $page->{DOM}->getElementsByTagName('link');
	push @links, $page->{DOM}->getElementsByTagName('a');
	push @links, $page->{DOM}->getElementsByTagName('area');

	foreach my $link (@links)
	{
		next unless ($link->getAttribute('rel') =~ /$regexp/i
			|| $link->getAttribute('rev') =~ /$regexp/i);
		my $href = $page->abs_url($link->getAttribute('href'));
		my $needToMergeObject;

		unless ($representative)
		{
			$representative = Cognition::RDFModel::Subject->new(
				Cognition::HTMLParser::Utils::BNodeURI(
					$page->{DOM}->documentElement,
					'RepresentativeContact'
				)
			);
			$representative->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
				Cognition::RDFModel::Value->new('http://xmlns.com/foaf/0.1/Person', 1));
			$representative->addVal('http://xmlns.com/foaf/0.1/page',
				Cognition::RDFModel::Value->new($pageURL, 1));
			$pageSubject->addVal('http://purl.org/uF/hCard/terms/representative',
				Cognition::RDFModel::Value->new($representative->uri, 1));
			$needToMergeSubject = 1;
		}
		
		# XFN to FOAF: Find Object of link.
		#
		# 1. If the link element is a descendant of an element with class="vcard"
		# which is not the representative hCard for the page, then this hCard
		# is taken to represent the person who is the object of the relationship.
		my $linkxp = $link->getAttribute('_xpath');
		my $bestlength = 99_999;
		my $best;
		my $friend;
		foreach my $hcard (@{$page->{uF}->{hCard}})
		{
			my $hcxp = $hcard->{_dom}->getAttribute('_xpath');
			
			if ((substr($linkxp, 0, length $hcxp) eq $hcxp)
			&&  (length $hcxp < $bestlength))
			{
				$bestlength = length $hcxp;
				$best = $hcard;
			}
		}
		$friend = findSubject($page->{RDF}, $best->uri)
			if ($best);
		
		# 2. If a non-representative hCard exists on the page with a UID property
		# which is identical to the target of the link element, then this hCard is
		# taken to represent the object of the relationship.
		$friend = findSubject($page->{RDF}, $href)
			unless ($friend);
		
		# 4. If no hCard representing the object of the relationship has been
		# found, then the object is represented by the following triples (note
		# that foaf:name is the link text)
		unless ($friend)
		{
			my $string = Cognition::HTMLParser::Utils::STRINGIFY($link);
			$string = $link->getAttribute('title') unless ($string);
			
			$friend = Cognition::RDFModel::Subject->new(
				Cognition::HTMLParser::Utils::BNodeURI($link, 'Friend') );
			$friend->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
				Cognition::RDFModel::Value->new('http://xmlns.com/foaf/0.1/Person', 1));
			$friend->addVal('http://xmlns.com/foaf/0.1/name',
				Cognition::RDFModel::Value->new($string));

			if ($href =~ /^mailto\:/i)
			{
				$friend->addVal('http://xmlns.com/foaf/0.1/mbox',
					Cognition::RDFModel::Value->new($href, 1));
			}
			elsif ($href =~ /^urn\:sha1\:([a-f0-9]+)$/i)
			{
				$friend->addVal('http://xmlns.com/foaf/0.1/mbox_sha1sum',
					Cognition::RDFModel::Value->new($1, 1));
			}
			elsif ($link->getAttribute('type') =~ /^image\//i)
			{
				$friend->addVal('http://xmlns.com/foaf/0.1/img',
					Cognition::RDFModel::Value->new($1, 1));
			}
			else
			{
				$friend->addVal('http://xmlns.com/foaf/0.1/page',
					Cognition::RDFModel::Value->new($href, 1));
			}
			
			$needToMergeObject = 1;
		}
		
		# Rel
		my $rel = lc($link->getAttribute('rel'));
		$rel =~ s/(^\s+|\s+$)//g;
		$rel =~ s/\s+/ /g;
		my @rel = split / /, $rel;
		foreach my $rel (@rel)
		{
			next unless ($rel =~ /$regexp/);
			
			if ($rel eq 'me')
			{
				if ($href =~ /^mailto\:/i)
				{
					$representative->addVal('http://xmlns.com/foaf/0.1/mbox',
						Cognition::RDFModel::Value->new($href, 1));
				}
				elsif ($href =~ /^urn\:sha1\:([a-f0-9]+)$/i)
				{
					$representative->addVal('http://xmlns.com/foaf/0.1/mbox_sha1sum',
						Cognition::RDFModel::Value->new($1, 1));
				}
				elsif ($link->getAttribute('type') =~ /^image\//i)
				{
					$representative->addVal('http://xmlns.com/foaf/0.1/img',
						Cognition::RDFModel::Value->new($1, 1));
				}
				else
				{
					$representative->addVal('http://xmlns.com/foaf/0.1/page',
						Cognition::RDFModel::Value->new($href, 1));
				}

				$needToMergeObject = 0;
				next;
			}
			
			$representative->addVal("http://gmpg.org/xfn/11#$rel",
				Cognition::RDFModel::Value->new($friend->uri,1));
			$representative->addVal('http://xmlns.com/foaf/0.1/knows',
				Cognition::RDFModel::Value->new($friend->uri,1))
				if ($R{$rel} =~ /K/);
			$friend->addVal('http://xmlns.com/foaf/0.1/knows',
				Cognition::RDFModel::Value->new($representative->uri,1))
				if ($R{$rel} =~ /K/);
			$friend->addVal("http://gmpg.org/xfn/11#$rel",
				Cognition::RDFModel::Value->new($representative->uri,1))
				if ($R{$rel} =~ /^S/);
			$friend->addVal("http://gmpg.org/xfn/11#$1",
				Cognition::RDFModel::Value->new($representative->uri,1))
				if ($R{$rel} =~ /I\:(.*)$/);
		}

		# Rev
		my $rev = lc($link->getAttribute('rev'));
		$rev =~ s/(^\s+|\s+$)//g;
		$rev =~ s/\s+/ /g;
		my @rev = split / /, $rev;
		foreach my $rev (@rev)
		{
			next unless ($rev =~ /$regexp/);
			
			if ($rev eq 'me')
			{
				if ($href =~ /^mailto\:/i)
				{
					$representative->addVal('http://xmlns.com/foaf/0.1/mbox',
						Cognition::RDFModel::Value->new($href, 1));
				}
				elsif ($href =~ /^urn\:sha1\:([a-f0-9]+)$/i)
				{
					$representative->addVal('http://xmlns.com/foaf/0.1/mbox_sha1sum',
						Cognition::RDFModel::Value->new($1, 1));
				}
				elsif ($link->getAttribute('type') =~ /^image\//i)
				{
					$representative->addVal('http://xmlns.com/foaf/0.1/img',
						Cognition::RDFModel::Value->new($1, 1));
				}
				else
				{
					$representative->addVal('http://xmlns.com/foaf/0.1/page',
						Cognition::RDFModel::Value->new($href, 1));
				}
				$needToMergeObject = 0;
				next;
			}
			
			$friend->addVal("http://gmpg.org/xfn/11#$rev",
				Cognition::RDFModel::Value->new($representative->uri,1));
			$friend->addVal('http://xmlns.com/foaf/0.1/knows',
				Cognition::RDFModel::Value->new($representative->uri,1));
			$representative->addVal('http://xmlns.com/foaf/0.1/knows',
				Cognition::RDFModel::Value->new($friend->uri, 1));
			$representative->addVal("http://gmpg.org/xfn/11#$rev",
				Cognition::RDFModel::Value->new($friend->uri,1))
				if ($R{$rev} =~ /^S/);
			$representative->addVal("http://gmpg.org/xfn/11#$1",
				Cognition::RDFModel::Value->new($friend->uri,1))
				if ($R{$rev} =~ /I\:(.*)$/);
		}
		
		$page->mergeSubject($friend)
			if ($needToMergeObject);
	}

	$page->mergeSubject($representative)
		if ($needToMergeSubject);
		
	$page->mergeNS('xfn', 'http://gmpg.org/xfn/11#', 'XHTML Friends Network');
	$page->mergeNS('hcterms', 'http://purl.org/uF/hCard/terms/', 'hCard Terms');

}

1;
