#!/usr/bin/perl

######################################################################
package Cognition::uF::hCard;
######################################################################

use CGI::Util;
use Cognition::Export;
use Cognition::Export::Contact;
use Cognition::HTMLParser::Utils;
use Cognition::MagicString;
use Cognition::uF;
use Data::Dumper;
use XML::LibXML qw(:all);
use strict;


sub consume
{
	my $page        = shift;
	my $pageURL     = $page->abs_url;
	my $pageSubject = $page->{RDF}->{$pageURL};

	my %levels;
	my $with_rel_me;
	my $with_url_source;
	
	if ( $page->{Profiles}->{'http://www.w3.org/2006/03/hcard'}
	||   $page->{Profiles}->{'http://purl.org/uF/hCard/1.0/'}
	||   $page->{Profiles}->{'http://purl.org/uF/2008/03/'}
	||  !$page->{options}->{uf_strict_profiles})
	{	
		$page->{uF}->{hCard} = [ Cognition::uF::hCard::parse_all($page) ];
		foreach my $hcard (@{ $page->{uF}->{hCard} })
		{
			$page->mergeSubject( $hcard->rdf_subject );
			
			my @xpath_parts = split /\//, $hcard->{_dom}->getAttribute('_xpath');
			for (my $i=0; defined $xpath_parts[$i]; $i++) 
			{
				if ($xpath_parts[$i] =~ /^address/i)
				{
					$levels{$i} = $hcard
						unless (defined $levels{$i});
				}
			}
			
			if ($hcard->{'_has_relme'})
			{
				$with_rel_me = $hcard unless ($with_rel_me);
			}
			
			foreach my $u (@{$hcard->{url}})
			{
				if ($u eq $pageURL)
				{
					$with_url_source = $hcard unless ($with_url_source);
				}
			}
			
			foreach my $c (@{$hcard->{tel}})
				{ $page->mergeSubject($c->rdf_subject); }
			foreach my $c (@{$hcard->{email}})
				{ $page->mergeSubject($c->rdf_subject); }
			foreach my $c (@{$hcard->{org}})
				{ $page->mergeSubject($c->rdf_subject); }
			foreach my $c (@{$hcard->{impp}})
				{ $page->mergeSubject($c->rdf_subject); }
			foreach my $c (@{$hcard->{label}})
				{ $page->mergeSubject($c->rdf_subject); }
			foreach my $c (@{$hcard->{n}})
				{ $page->mergeSubject($c->rdf_subject); }
		}
		if (defined $page->{uF}->{hCard}->[0])
			{ $page->mergeNS( $page->{uF}->{hCard}->[0] ); }
	}
	
	# Determine "shallowest" <address> element: take this to be the contact
	# for the page.
	if (keys %levels)
	{
		my @sorted_levels = sort keys %levels;
		my $shallow_level = shift @sorted_levels;
		my $contact_hcard = $levels{$shallow_level};
		$pageSubject->addVal('http://purl.org/uF/hCard/terms/contact',
			Cognition::RDFModel::Value->new($contact_hcard->uri, 1));
	}
	
	unless (firstOf($pageSubject, 'http://purl.org/uF/hCard/terms/representative'))
	{		
		# Representative hCard
		my @topics = allOf($pageSubject, 'http://xmlns.com/foaf/0.1/primaryTopic');
		my $rep;
		if (@topics)
		{
			foreach my $topic (@topics)
			{
				my $topicSubject = findSubject($page->{RDF}, $topic);
				$rep = $topicSubject
					if ($topicSubject && Cognition::Export::Contact::is($topicSubject));
			}
		}
		$rep = $rep || $with_rel_me || $with_url_source || $page->{uF}->{hCard}->[0];
		if ($rep)
		{
			$pageSubject->addVal('http://purl.org/uF/hCard/terms/representative',
				Cognition::RDFModel::Value->new($rep->uri, 1));
		}
	}
	
}


sub parse_all
{
	my $page    = shift;
	my $within  = shift || $page->{DOM};
	my $isInCal = shift || 0;
	my @nodes   = searchClass('vcard', $within);
	my @rv;
	
	foreach my $card (@nodes)
	{
		my $hcard = parse($page, $card, $isInCal);
		push @rv, $hcard;
	}
	
	return @rv;
} #/sub parse_all


sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;
	
	if (length $this->{uid})
	{
		my $uid = $this->{uid};
		$uid = "urn:x-uid:$uid"
			unless ($uid =~ /^[a-z0-9]{1,10}\:/i);
		push @rv, $uid;
	}
	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	# if (length $this->{fn})
	# {
	# 	push @rv, 'ldap:///cn=' . CGI::Util::escape($this->{fn});
	# }
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_xpath}, 'Contact')
		unless (@rv);

	unless (defined $all) { return $rv[0];}
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri


sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:2426#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'Vcard', 1));
	
	# Patterns for simple data.
	my @_SINGLES  = qw(fn bday tz sort-string class dday gender kind
	                   cutype rsvp partstat member delegated-from sent-by);
	my @_SINGLESR = qw(uid);
	my @_MULTIS   = qw(nickname title role key mailer note rev lang);
	my @_MULTISR  = qw(url photo sound logo);
	my @_OBJECTS  = qw(category adr geo agent tel email org impp n label species
	                   member);
	my @_1OBJECTS = qw(birth death);
	
	# Properties which exist as singles.
	foreach my $property (@_SINGLES)
	{
		next unless (length $this->{$property});
		my $myns = $ns;
		if ($property =~ /^(cutype|rsvp|partstat|member|delegated.from|sent.by)$/)
			{ $myns = 'urn:ietf:rfc:2445#'; }
		my $val = Cognition::RDFModel::Value->new($this->{$property});
		$subject->addVal($myns.$property, $val);
	}

	# Resource-type properties which exist as singles.
	foreach my $property (@_SINGLESR)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property}, 1);
		$subject->addVal($ns.$property, $val);
	}
		
	# Properties which exist as multiple values.
	foreach my $property (@_MULTIS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V);
			$subject->addVal($ns.$property, $val);
		}
	}
	
	# Resource-type properties which exist as multiple values.
	foreach my $property (@_MULTISR)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V, 1);
			$subject->addVal($ns.$property, $val);
		}
	}

	# Embedded objects which may be represented as resources
	foreach my $property (@_OBJECTS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			my $val;			
			if (UNIVERSAL::can($V, 'uri'))
				{ $val = Cognition::RDFModel::Value->new($V->uri, 1); }
			elsif ($property eq 'member')
				{ $val = Cognition::RDFModel::Value->new($V, 1); }
			else
				{ $val = Cognition::RDFModel::Value->new($V); }
				
			if ($property eq 'species')
				{ $subject->addVal('http://purl.org/NET/biol/ns#hasTaxonomy', $val); }
			elsif ($property eq 'impp')
				{ $subject->addVal('urn:ietf:rfc:4770#impp', $val); }
			else
				{ $subject->addVal($ns.$property, $val); }
		}
	}
	
	# Embedded objects which may be represented as resources
	foreach my $property (@_1OBJECTS)
	{
		if ( $this->{$property} )
		{
			my $V = $this->{$property};
			my $val;			
			if (UNIVERSAL::can($V, 'uri'))
				{ $val = Cognition::RDFModel::Value->new($V->uri, 1); }
			else
				{ $val = Cognition::RDFModel::Value->new($V); }
				
			$subject->addVal($ns.$property, $val);
		}
	}
	
	foreach my $m (@{ $this->{measures} })
	{
		$subject->addVal('http://buzzword.org.uk/rdf/measure-aux#hasMeasurement',
			Cognition::RDFModel::Value->new($m->uri, 1));
		$subject->addVal($m->dim_uri,
			Cognition::RDFModel::Value->new($m->qv_uri, 1))
			if (length $m->dim_uri);
	}
	
	return $subject;
	
} #/sub rdf_subject


sub metadata_ns
{
	my $this = shift;
	return {
		'vcard'     => { nsurl=>'urn:ietf:rfc:2426#',title=>'vCard' },
		'vcardIMPP' => { nsurl=>'urn:ietf:rfc:4770#',title=>'vCard IMPP Extension' },
		'hcterms'   => { nsurl=>'http://purl.org/uF/hCard/terms/',title=>'hCard Terms'}
		};
} #/sub namespace


# TODO: Allow birth and death to be formatted addresses rather than plain
#       old strings.
sub parse
{
	my $page       = shift;
	my $rv         = { '_dom'=>shift };
	my $isInCal    = shift || 0;
	my $hcard_node = $rv->{'_dom'}->cloneNode(1);
	
	my $id = $hcard_node->getAttribute('id');
	$rv->{_id} = $id if (length $id);
	$rv->{_page} = $page;
	$rv->{_xpath} = $rv->{_dom}->getAttribute('_xpath');
		
	Cognition::uF::data_patterns($page, $hcard_node, 6);
	
	my $properties = [
		['adr',         'm*',   {'embedded'=>'adr'}],
		['agent',       'M*',   {'embedded'=>'hCard'}],
		['bday',        'd?'],
		['biota',       'm*',   {'embedded'=>'species', 'use-key'=>'species'}], #extension
		['birth',       'M?',   {'embedded'=>'hCard adr geo'}], #extension
		['category',    '*'],
		['class',       '?'],
		['dday',        'd?'],  #extension
		['death',       'M?',   {'embedded'=>'hCard adr geo'}], #extension
		['fn',          '1'],
		['gender',      '?'],   #extension
		['geo',         'm*',   {'embedded'=>'geo'}],
		['kind',        '?'],   #extension
		['key',         'u*'],
		['lang',        '*'],   #extension
		['logo',        'u*'],
		['mailer',      '*'],
		['member',      'Mu*',  {'embedded'=>'hCard'}], #extension
		['nickname',    '*'],
		['note',        '*'],
		['photo',       'u*'],
		['rev',         'd*'],
		['role',        '*'],
		['sort-string', '?'],
		['sound',       'u*'],
		['title',       '*'],
		['tz',          '?'],
		['uid',         'U?'],
		['url',         'u*']
	];
	push @{$properties}, (
		['cn',          '?'],
		['cutype',      '?'],
		['member',      '?'],
		['rsvp',        '?'],
		['delegated-from', 'Mu?', {'embedded'=>'hCard'}],
		['sent-by',     'Mu?',  {'embedded'=>'hCard'}],
	) if ($isInCal);

	simple_parse(
		$rv,
		$hcard_node,
		$properties,
		{
			'rel-me'     => '_has_rel_me',
			'rel-tag'    => 'category',
			'hmeasure'   => 'measures', #extension
			'no-destroy' => ['adr', 'geo'],
			'my-root'    => 'vcard'
		}
	);

	# Names
	my @names = searchClass('n', $hcard_node);
	@names = ($hcard_node) unless (@names);
	foreach my $name (@names)
	{
		my $n = Cognition::uF::hCard::Name::parse($page, $name);
		if ($n->{_n})
		{
			$n->{_hcard} = $rv;
			push @{ $rv->{n} }, $n;
			last;
		}
	}
	my $fnNode;
	my @names = searchClass('fn', $hcard_node);
	$fnNode = $names[0];

	# Parse organisations - TODO: clean this up using simple_parse?
	my @orgs = searchClass('org', $hcard_node);
	foreach my $org (@orgs)
	{
		my @orgnames = searchClass('organization-name', $org);
		my $parsed_org = {};
		
		if (!@orgnames)
		{
			$parsed_org = { 
				'organization-unit' => undef,
				'organization-name' => STRINGIFY($org, 'value'),
			};
			$parsed_org->{'_id'}               = $org->getAttribute('id');
			$parsed_org->{'_page'}             = $page;
			$parsed_org->{'_hcard'}            = $rv;
			$parsed_org->{'_dom'}              = $org;
		}
		else
		{
			my @orgunits = searchClass('organization-unit', $org);
			my @vat      = searchClass('(vat-number|x-vat-number)', $org);
			my @co       = searchClass('(company-number|x-company-number)', $org);
			my @ch       = searchClass('(charity-number|x-charity-number)', $org);

			$parsed_org = {};
			$parsed_org->{'organization-name'} = STRINGIFY($orgnames[0], 'value');
			$parsed_org->{'organization-unit'} = ();
			$parsed_org->{'x-vat-number'}      = STRINGIFY($vat[0], 'value');
			$parsed_org->{'x-company-number'}  = STRINGIFY($co[0], 'value');
			$parsed_org->{'x-charity-number'}  = STRINGIFY($ch[0], 'value');
			$parsed_org->{'_id'}               = $org->getAttribute('id');
			$parsed_org->{'_page'}             = $page;
			$parsed_org->{'_hcard'}            = $rv;
			$parsed_org->{'_dom'}              = $org;
			foreach my $ou ( @orgunits )
			{
				push @{ $parsed_org->{'organization-unit'} }, STRINGIFY($ou, 'value');
			}
		}
		bless $parsed_org, 'Cognition::uF::hCard::org';
		push @{ $rv->{org} }, $parsed_org;
	}
	if (!@orgs)
	{
		my @orgnames = searchClass('organization-name', $hcard_node);
		my @orgunits = searchClass('organization-unit', $hcard_node);
		my @vat = searchClass('(vat-number|x-vat-number)', $hcard_node);
		my @co  = searchClass('(company-number|x-company-number)', $hcard_node);
		my @ch  = searchClass('(charity-number|x-charity-number)', $hcard_node);

		my $parsed_org = {};
		$parsed_org->{'organization-name'} = STRINGIFY($orgnames[0], 'value');
		$parsed_org->{'organization-unit'} = ();
		$parsed_org->{'x-vat-number'}      = STRINGIFY($vat[0], 'value');
		$parsed_org->{'x-company-number'}  = STRINGIFY($co[0], 'value');
		$parsed_org->{'x-charity-number'}  = STRINGIFY($ch[0], 'value');
		$parsed_org->{'_page'}             = $page;
		$parsed_org->{'_hcard'}            = $rv;
		$parsed_org->{'_implied'}          = '1';
		$parsed_org->{'_dom'}              = $hcard_node;

		foreach my $ou ( @orgunits )
		{
			push @{ $parsed_org->{'organization-unit'} }, STRINGIFY($ou, 'value');
		}
		bless $parsed_org, 'Cognition::uF::hCard::org';
		push @{ $rv->{org} }, $parsed_org if ($parsed_org->{'organization-name'}
			|| $parsed_org->{'organization-unit'}->[0]);
	}	
	
	# Telephone numbers and e-mail addresses
	my @comms = ('tel', 'email', 'impp', 'label');
	foreach my $comm (@comms)
	{
		my @nodes = searchClass($comm, $hcard_node);
		foreach my $n (@nodes)
		{
			my @accumValue;
			my @accumType;
			
			foreach my $v (searchClass('value', $n))
			{
				if ($comm eq 'label')
				{
					push @accumValue, STRINGIFY($v, undef, 1);
				}
				elsif ($v->localname eq 'a' || $v->localname eq 'area')
				{
					my $href = $v->getAttribute('href');

					push @accumType, 'fax' if ($href =~ /^fax\:/i);
					push @accumType, 'modem' if ($href =~ /^modem\:/i);

					$href =~ s/(^[a-z]+\:|\?.*$)//g unless ($comm eq 'impp');

					push @accumValue, $href;
				}
				elsif ($v)
				{
					push @accumValue, STRINGIFY($v, undef, 1);
				}
			}
			
			foreach my $v (searchClass('type', $n))
			{
				push @accumType, uc(STRINGIFY($v, undef, 1));
				$v->getParentNode->removeChild($v);
			}
			
			my $val = join '', @accumValue;
			if ($val !~ /[0-9a-z]/i)
			{
				if ($n->localname eq 'a' || $n->localname eq 'area')
				{
					$val = $n->getAttribute('href');
					$val =~ s/(^[a-z]+\:|\?.*$)//g;
				}
				else
				{
					$val = STRINGIFY($n, undef, 1);
				}
			}
			if ($comm eq 'tel')
			{
				$val = lc($val);
				$val =~ s/[^0-9x+-]//g;
			}
			
			my $C = {
					'value'  => $val,
					'type'   => \@accumType,
					'_id'    => $n->getAttribute('id'),
					'_page'  => $page,
					'_hcard' => $rv,
					'_dom'   => $n
				};
			bless $C, "Cognition::uF::hCard::$comm";
			push @{ $rv->{$comm} }, $C if (length $val);
		}
		
	}
	
	$rv->{fn} = $rv->{cn}
		unless ($rv->{fn});
	
	# Check to see if this contact card is for an individual contact or an org.
	# If for an org, disable n-optimisation.
	if ($rv->{kind})
	{
		$rv->{kind} =~ s/(^\s|\s+$)//g;
		$rv->{kind} = lc($rv->{kind});
	}
	elsif (length $rv->{fn})
	{
		$rv->{kind} = 'individual';
		
		ORGLOOP: foreach my $org (@{ $rv->{org} })
		{
			if ("".$org->{'organization-name'} eq $rv->{fn})
			{
				$rv->{kind} = 'org';
				last ORGLOOP;
			}
			foreach my $ou (@{ $org->{'organization-unit'} })
			{
				if ("$ou" eq $rv->{fn})
				{
					$rv->{kind} = 'group';
					last ORGLOOP;
				}
			}
		}
		
		if ($rv->{kind} eq 'individual')
		{
			ADRLOOP: foreach my $adr (@{ $rv->{adr} })
			{
				foreach my $part (qw(post-office-box extended-address
					street-address locality region postal-code country-name))
				{
					foreach my $line (@{ $adr->{$part} })
					{
						if ("$line" eq $rv->{fn})
						{
							$rv->{kind} = 'x-'.$part;
							last ADRLOOP;
						}
					}
				}
			}
		}
	}

	# Perform n-optimisation.
	if ($rv->{kind} eq 'individual')
	{
		unless ($rv->{n})
		{
			$rv->{fn} =~ s/(^\s|\s$)//g;
			$rv->{fn} =~ s/\s+/ /g;
			
			my @words = split / /, $rv->{fn};
			
			if (scalar @words == 1)
			{
				$rv->{nickname} = [ $words[0] ];
			}
			elsif (scalar @words)
			{
				if (($words[0] =~ /^.*\,$/ || $words[1] =~ /^.\.?$/) && !defined $words[2])
				{
					$words[0] =~ s/[\.\,]$//;
					$words[1] =~ s/[\.\,]$//;
					$rv->{n}->[0]->{'given-name'}->[0] = $words[1];
					$rv->{n}->[0]->{'family-name'}->[0] = $words[0];
					$rv->{n}->[0]->{_dom} = $fnNode;
					bless $rv->{n}->[0], 'Cognition::uF::hCard::Name';
				}
				elsif (!defined $words[2])
				{
					$rv->{n}->[0]->{'given-name'}->[0] = $words[0];
					$rv->{n}->[0]->{'family-name'}->[0] = $words[1];
					$rv->{n}->[0]->{_dom} = $fnNode;
					bless $rv->{n}->[0], 'Cognition::uF::hCard::Name';
				}
			}
		}
	}

	bless $rv;
	return $rv;
	
} #/sub parse

sub validate
{
	my $this = shift;
	my @rv;
	
	unless (length $this->{fn})
	{
		push @rv, ['hCard', $this->uri, 'E', 1, "No formatted name (fn).", "The formatted name property is required."];
	}
	
	foreach my $u (@{ $this->{url} })
	{
		my $U = URI->new($u);
		unless (length $U->scheme)
		{
			push @rv, ['hCard', $this->uri, 'E', 11, "URL property does not conform to IRI syntax.", "The URL property must be a valid address for a website or another type of IRI.", $u];
		}
	}
	
	if ($this->{uid})
	{
		my $U = URI->new($this->{uid});
		unless (length $U->scheme)
		{
			push @rv, ['hCard', $this->uri, 'W', 101, "UID property does not conform to IRI syntax.", "The UID property should be a valid address for a website or another type of IRI."];
		}
	}

	foreach my $prop (qw(photo logo sound))
	{
		my $n = ($prop eq 'photo')?2:(($prop eq 'logo')?3:4);
		foreach my $v (@{$this->{$prop}})
		{
			my $U = URI->new($v);
			unless (length $U->scheme)
			{
				push @rv, ['hCard', $this->uri, 'E', $n, uc($prop)." property does not conform to IRI syntax.", "This should be a resolvable URL or another type of IRI.", $v];
			}
		}
	}
	
	foreach my $t (@{$this->{tel}})
		{ push @rv, $t->validate; }
	foreach my $t (@{$this->{email}})
		{ push @rv, $t->validate; }
	foreach my $t (@{$this->{adr}})
		{ push @rv, $t->validate; }
	foreach my $t (@{$this->{org}})
		{ push @rv, $t->validate; }
	foreach my $t (@{$this->{geo}})
		{ push @rv, $t->validate; }
	foreach my $t (@{$this->{label}})
		{ push @rv, $t->validate; }
	foreach my $t (@{$this->{n}})
		{ push @rv, $t->validate; }
	
	return @rv;
}

1; #/package Cognition::uF::hCard



######################################################################
package Cognition::uF::hCard::tel;
######################################################################

sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;

#	if (length $this->{value})
#	{
#		my $v = $this->{value};
#		if ($v =~ /x/)
#		{
#			$v =~ s/x/;ext=/;
#		}
#		if ($v !~ /^\+/)
#		{
#			$v .= ';phone-context=unknown-context.invalid';
#		}
#		$v .= ';x-hcard-context=' . CGI::Util::escape($this->{_hcard}->uri)
#			if (defined $this->{_hcard});
#		push @rv, "tel:$v";
#	}
	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_dom}, 'ContactPhone')
		unless (@rv);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri

sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:2426#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'Tel', 1));
	
	# Patterns for simple data.
	my @_SINGLES  = qw(value);
	my @_MULTIS   = qw(type);
	
	# Properties which exist as singles.
	foreach my $property (@_SINGLES)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property});
		$subject->addVal($ns.$property, $val);
	}

	# Properties which exist as multiple values.
	foreach my $property (@_MULTIS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V);
			$subject->addVal($ns.$property, $val);
		}
	}
	
	return $subject;
	
} #/sub rdf_subject

sub validate
{
	my $this = shift;
	my @rv;
	
	foreach my $t (@{$this->{type}})
	{
		if ($t !~ /^\s*(voice|home|msg|work|pref|fax|cell|video|pager|bbs|modem
			|car|isdn|pcs|x\-.+)\s*$/ix )
		{
			push @rv, ['hCard', $this->{'_hcard'}->uri, 'E', 21, "Invalid telephone number type.", "Valid telephone number types are: VOICE, home, msg, work, pref, fax, cell, video, pager, bbs, modem, car, isdn, pcs. If a non-standard type is required, it must be prefixed with 'X-'.", $t];		
		}
	}

	if ($this->{value} !~ /^\+/)
	{
		push @rv, ['hCard', $this->{'_hcard'}->uri, 'W', 121, "Local phone number.", "On the Web, international phone numbers (prefixed with a + sign) are more useful.", $this->{value}];		
	}
	
	return @rv;
	
} #/sub validate

1; #/package Cognition::uF::hCard::tel





######################################################################
package Cognition::uF::hCard::label;
######################################################################

sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;

	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_dom}, 'ContactAddressLabel')
		unless (@rv);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri

sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:2426#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'Label', 1));
	
	# Patterns for simple data.
	my @_SINGLES  = qw(value);
	my @_MULTIS   = qw(type);
	
	# Properties which exist as singles.
	foreach my $property (@_SINGLES)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property});
		$subject->addVal($ns.$property, $val);
	}

	# Properties which exist as multiple values.
	foreach my $property (@_MULTIS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V);
			$subject->addVal($ns.$property, $val);
		}
	}
	
	return $subject;
	
} #/sub rdf_subject

sub validate
{
	my $this = shift;
	my @rv;
	
	foreach my $t (@{$this->{type}})
	{
		if ($t !~ /^\s*(INTL|POSTAL|PARCEL|WORK|dom|home|pref|x\-.+)\s*$/ix )
		{
			push @rv, ['hCard', $this->{'_hcard'}->uri, 'E', 61, "Invalid label type.", "Valid label types are: INTL, POSTAL, PARCEL, WORK, dom, home and pref. If a non-standard type is required, it must be prefixed with 'X-'.", $t];		
		}
	}

	return @rv;
	
} #/sub validate

1; #/package Cognition::uF::hCard::label





######################################################################
package Cognition::uF::hCard::email;
######################################################################

sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;

#	if (length $this->{value})
#	{
#		my $v = $this->{value};
#		$v .= '?x-hcard-context=' . CGI::Util::escape($this->{_hcard}->uri)
#			if (defined $this->{_hcard});
#		push @rv, "mailto:$v";
#	}
	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_dom}, 'ContactEmail')
		unless (@rv);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri

sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:2426#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'Email', 1));
	
	# Patterns for simple data.
	my @_SINGLES  = qw(value);
	my @_MULTIS   = qw(type);
	
	# Properties which exist as singles.
	foreach my $property (@_SINGLES)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property});
		$subject->addVal($ns.$property, $val);
	}

	# Properties which exist as multiple values.
	foreach my $property (@_MULTIS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V);
			$subject->addVal($ns.$property, $val);
		}
	}
	
	return $subject;
	
} #/sub rdf_subject

sub validate
{
	my $this = shift;
	my @rv;
	
	foreach my $t (@{$this->{type}})
	{
		if ($t !~ /^\s*(INTERNET|x400|pref|x\-.+)\s*$/ix)
		{
			push @rv, ['hCard', $this->{'_hcard'}->uri, 'E', 31, "Invalid e-mail type.", "Valid e-mail types are: internet, x400, pref. If a non-standard type is required, it must be prefixed with 'X-'.", $t];		
		}
	}

	if ($this->{value} !~ /\@/)
	{
		push @rv, ['hCard', $this->{'_hcard'}->uri, 'W', 131, "Non-RFC822 e-mail address.", "On the Web, standard RFC 822 e-mail addresses are more useful.", $this->{value}];		
	}
	
	return @rv;
	
} #/sub validate

1; #/package Cognition::uF::hCard::email


######################################################################
package Cognition::uF::hCard::impp;
######################################################################

sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;

#	if (length $this->{value})
#	{
#		my $v = $this->{value};
#		if (defined $this->{_hcard})
#		{
#			$v .= ($v =~ /\?/) ? '&' : '?';
#			$v .= 'x-hcard-context=' . CGI::Util::escape($this->{_hcard}->uri);
#		}
#		push @rv, $v;
#	}
	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_dom}, 'ContactChat')
		unless (@rv);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri

sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:4770#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'Impp', 1));
	
	# Patterns for simple data.
	my @_SINGLESR = qw(value);
	my @_MULTIS   = qw(type);
	
	# Resource-type properties which exist as singles.
	foreach my $property (@_SINGLESR)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property}, 1);
		$subject->addVal($ns.$property, $val);
	}
		
	# Properties which exist as multiple values.
	foreach my $property (@_MULTIS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V);
			$subject->addVal($ns.$property, $val);
		}
	}
	
	return $subject;
	
} #/sub rdf_subject

sub validate
{
	my $this = shift;
	my @rv;
	
	return @rv;
	
} #/sub validate

1; #/package Cognition::uF::hCard::impp


######################################################################
package Cognition::uF::hCard::org;
######################################################################

sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;

#	if (length $this->{'organization-name'})
#	{
#		my $v = 'o='.CGI::Util::escape($this->{'organization-name'});
#		foreach my $ou ( @{$this->{'organization-unit'}} )
#		{
#			$v .= ',ou='.CGI::Util::escape($ou);
#		}
#		push @rv, "ldap:///$v";
#	}
	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_dom}, 'Org')
		unless (@rv);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri

sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:2426#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'Org', 1));
	
	# Patterns for simple data.
	my @_SINGLES  = qw(organization-name x-charity-number x-vat-number
		x-company-number);
	my @_MULTIS   = qw(organization-unit);
	
	# Properties which exist as singles.
	foreach my $property (@_SINGLES)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property});
		$subject->addVal($ns.$property, $val);
	}

	# Properties which exist as multiple values.
	foreach my $property (@_MULTIS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V);
			$subject->addVal($ns.$property, $val);
		}
	}
	
	return $subject;
	
} #/sub rdf_subject


sub validate
{
	my $this = shift;
	my @rv;
	
	if ($this->{implied})
	{
		push @rv, ['hCard', $this->{'_hcard'}->uri, 'E', 51, "Organisation subtypes found without an 'org' wrapper.", "The 'organization-name' and 'organization-unit' properties are only valid within an 'org'.", $this->{'organization-name'}];		
	}
	
	return @rv;
	
} #/sub validate


1; #/package Cognition::uF::hCard::org




######################################################################
package Cognition::uF::hCard::Name;
######################################################################

use Cognition::HTMLParser::Utils;
use Cognition::uF;

sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;

	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_dom}, 'ContactName')
		unless (@rv);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri

sub parse
{
	my $page = shift;
	my $dom  = shift;
	my $rv   = {};
	
	simple_parse(
		$rv,
		$dom,
		[
			['additional-name',  '*'],
			['family-name',      '*'],
			['given-name',       '*'],
			['honorific-prefix', '*'],
			['honorific-suffix', '*'],
			['initial',          '*'], # extension
		],
		{
			'no-destroy' => ['adr', 'geo']
		}
	);

	$rv->{_id}   = $dom->getAttribute('id') if (length $dom->getAttribute('id'));
	$rv->{_dom}  = $dom;
	$rv->{_page} = $page;
	
	bless $rv;
}

sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:2426#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'Name', 1));
	
	# Patterns for simple data.
	my @_MULTIS   = qw(family-name given-name additional-name honorific-prefix
	                   honorific-suffix initial);
	
	# Properties which exist as multiple values.
	foreach my $property (@_MULTIS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V);
			$subject->addVal($ns.$property, $val);
		}
	}
	
	my $value = '';
	foreach my $prop (qw(family-name given-name additional-name honorific-prefix
	                   honorific-suffix))
	{
		my $v = join 'XCOMMAX', @{ $this->{$prop} };
		$v =~ s/\\/\\\\/g;
		$v =~ s/\,/\\\,/g;
		$v =~ s/\;/\\\;/g;
		$v =~ s/XCOMMAX/\,/g;
		
		$value .= $v.';';
	}
	$value =~ s/\;+$//g;
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#value',
		Cognition::RDFModel::Value->new($value))
		if (length $value);
	
	$value = '';
	foreach my $prop (qw(honorific-prefix given-name additional-name family-name
	                   honorific-suffix))
	{
		my $v = join ' ', @{ $this->{$prop} };
		$value .= $v.' ' if (length $v);
	}
	$value =~ s/\s+$//g;
	
	$subject->addVal('http://www.w3.org/2000/01/rdf-schema#label',
		Cognition::RDFModel::Value->new($value))
		if (length $value);
	
	return $subject;
	
} #/sub rdf_subject


sub validate
{
	my $this = shift;
	my @rv;
	
	return @rv;
	
} #/sub validate



1; #/package Cognition::uF::hCard::Name
