#!/usr/bin/perl

######################################################################
package Cognition::uF::hEntry;
######################################################################

use CGI::Util;
use Cognition::HTMLParser::Utils;
use Cognition::uF;
use XML::LibXML qw(:all);
use strict;

sub parse_all
{
	my $page   = shift;
	my $within = shift || $page->{DOM};
	my @nodes1 = searchClass('hentry', $within);
	my @nodes2 = searchClass('hslice', $within);
	my @nodes = (@nodes1, @nodes2);
	my @rv;
	
	foreach my $e (@nodes)
	{
		my $E = parse($page, $e);
		push @rv, $E;
	}
	
	return @rv;
} #/sub parse_all


sub uri
{
	my $this = shift;
	my $all  = shift;
	my @rv;

	if (length $this->{_id})
	{
		push @rv, $this->{_page}->abs_url.'#'.$this->{_id};
	}
	push @rv, Cognition::HTMLParser::Utils::BNodeURI($this->{_dom}, 'FeedEntry')
		unless (@rv);
	
	return $rv[0] unless (defined $all);
	while ($all) { shift @rv; $all--; }
	return @rv;
} #/sub uri


sub rdf_subject
{
	my $this    = shift;
	my $ns      = 'urn:ietf:rfc:4287#';

	# Create RDF subject
	my $subject = Cognition::RDFModel::Subject->new($this->uri);
	$subject->addObject($this);
	$subject->setDom($this->{_dom});
	
	$subject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
		Cognition::RDFModel::Value->new($ns.'entry', 1));
	
	# Patterns for simple data.
	my @_SINGLES  = qw(title content summary published);
	my @_SINGLESR = qw(link parent);
	my @_MULTIS   = qw(updated);
	my @_MULTISR  = qw(child sibling);
	my @_OBJECTS  = qw(author category enclosure);
	
	# Properties which exist as singles.
	foreach my $property (@_SINGLES)
	{
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property});
		$subject->addVal($ns.$property, $val);
	}

	# Resource-type properties which exist as singles.
	foreach my $property (@_SINGLESR)
	{
		my $myns = $ns;
		$myns = 'urn:ietf:rfc:2445#' if ($property eq 'parent');
		next unless (length $this->{$property});
		my $val = Cognition::RDFModel::Value->new($this->{$property}, 1);
		$subject->addVal($myns.$property, $val);
	}
		
	# Properties which exist as multiple values.
	foreach my $property (@_MULTIS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V);
			$subject->addVal($ns.$property, $val);
		}
	}
	
	# Resource-type properties which exist as multiple values.
	foreach my $property (@_MULTISR)
	{
		foreach my $V (@{ $this->{$property} })
		{
			my $myns = $ns;
			$myns = 'urn:ietf:rfc:2445#' if ($property eq 'child' || $property eq 'sibling');
			next unless (length $V);
			my $val = Cognition::RDFModel::Value->new($V, 1);
			$subject->addVal($myns.$property, $val);
		}
	}

	# Embedded objects which may be represented as resources
	foreach my $property (@_OBJECTS)
	{
		foreach my $V (@{ $this->{$property} })
		{
			my $val;
			if (ref $V)
				{ $val = Cognition::RDFModel::Value->new($V->uri, 1); }
			else
				{ $val = Cognition::RDFModel::Value->new($V); }
			$subject->addVal($ns.$property, $val);
		}
	}
	
	return $subject;
	
} #/sub rdf_subject


sub metadata_ns
{
	my $this = shift;
	return { 'atom' => { nsurl=>'urn:ietf:rfc:4287#',title=>'Atom' } };
} #/sub namespace


sub parse
{
	my $page = shift;
	my $rv = { '_dom'=>shift };
	bless $rv;
	
	# hAtom spec: an Entry element is identified by class name hentry
	my $root_node = $rv->{'_dom'}->cloneNode(1);
	
	my $id = $root_node->getAttribute('id');
	$rv->{_id} = $id if (length $id);
	$rv->{_page} = $page;
	
	data_patterns($page, $root_node, 6);

	# Author - too complex for simple_parse().
	my @nested = Cognition::HTMLParser::Utils::searchClass('vcard', $root_node);
	foreach my $h (@nested)
	{
		if ($h->getAttribute('class') =~ / (^|\s) author (\s|$) /x)
		{
			push @{ $rv->{author} }, Cognition::uF::hCard::parse($page, $h);
		}
	}	
	if (!defined $rv->{author})
	{
		foreach my $h (@nested)
		{
			if ($h->tagName == 'address')
			{
				push @{ $rv->{author} }, Cognition::uF::hCard::parse($page, $h);
			}
		}
		if (!defined $rv->{author})
		{
			foreach my $address ($root_node->getElementsByTagName('address'))
			{
				push @{ $rv->{author} }, Cognition::uF::hCard::parse_all($page, $address);
			}
		}
	}
	
	simple_parse(
		$rv,
		$root_node,
		[
			['bookmark',        'ru?', {'use-key'=>'link'}],
			['entry-content',   '&',   {'use-key'=>'content'}],
			['entry-summary',   '&',   {'use-key'=>'summary'}],
			['published',       'd?'],
			['updated',         'd*']
		],
		{
			'rel-tag'       => 'category',
			'rel-enclosure' => 'enclosure',
			'rel-license'   => 'license'   # TODO: RDF bit
		}
	);
	
	# Entry title - has weird fallback rules.
	my @nodes = Cognition::HTMLParser::Utils::searchClass('entry-title', $root_node);
	if (@nodes)
	{
		$rv->{title} = STRINGIFY($nodes[0], 'value');
	}
	else
	{
		foreach my $tag ($root_node->getElementsByTagName('*'))
		{
			if ($tag->tagName =~ /^h[1-9]?$/i)
			{
				$rv->{title} = STRINGIFY($tag, 'value');
				last;
			}
		}
	}
	if (!defined $rv->{title})
	{
		# TODO
		# Should fetch <title> here from $page unless $page contains an element
		# of class 'hfeed'.
	}
		
	# Bookmark - fallback rule
	unless (length $rv->{link})
	{
		$rv->{link} = $rv->uri;
	}
	
	return $rv;
	
} #/sub parse


1; #/package Cognition::uF::hEntry

